# VICE package.
#
# Exports class Vice.
# Usage:
# $v = new Vice([voice, [certificate]]);
# $v->load( servername );
# $v->clear()
# $request = $v->spawn{'request'};
# $fail = $v->transid($request);
# $fail = $v->pickup() unless $fail;
# $fail = $v->save_cert(cert_file, [overwrite]);
# $v->save()

package Vice;

BEGIN {
    $VICE_support = '<A Href="http://www.verisign.com/support/">VeriSign support</A>';

    # VICE error codes
    %VICE_error_codes =
        ( 'NO_ERROR' => 0x0000,
          'TE_ALLOC' => 0x0100,
          'TE_ATTRIBUTE_TAG' => 0x0105,
          'TE_ATTRIBUTE_TYPE' => 0x0106,
          'TE_ATTRIBUTE_VALUE' => 0x0107,
          'TE_ATTRIBUTE_VALUE_LEN' => 0x0108,
          'TE_BER_ENCODING' => 0x010a,
          'TE_CERT_CHAIN' => 0x010c,
          'TE_CERT_ENCODING' => 0x010d,
          'TE_CRL_ENCODING' => 0x0110,
          'TE_DEK_ALG_NOT_SUPPORTED' => 0x0114,
          'TE_DEK_ALG_UNKNOWN' => 0x0115,
          'TE_IO' => 0x0120,
          'TE_MIC_ALG_NOT_SUPPORTED' => 0x0125,
          'TE_MIC_ALG_UNKNOWN' => 0x0126,
          'TE_NAME_OBJ' => 0x0128,
          'TE_NOT_FOUND' => 0x012a,
          'TE_PBE_ALG_NOT_SUPPORTED' => 0x0132,
          'TE_PBE_ALG_UNKNOWN' => 0x0133,
          'TE_SIGNATURE_ALG_NOT_SUPPORTED' => 0x013c,
          'TE_SIGNATURE_ALG_UNKNOWN' => 0x013d,
          'TE_INVALID_AUTHORITY_KEY_ID' => 0x0145,
          'TE_INVALID_KEY_USAGE' => 0x0146,
          'TE_INVALID_BASIC_CONSTRAINT' => 0x0147,
          'TE_INVALID_PRIVATE_KEY_USAGE' => 0x0148,
          'TE_INVALID_SUBJECT_ALT_NAME' => 0x014a,
          'eCMS_OVERLAPPING_USER_CERT' => 0x1005,
          'eECAS_PRODUCT_NOT_FOUND' => 0x3041,
          'eECAS_UNKNOWN_PRODUCT_CFG_ERROR' => 0x3047,
          'eECAS_ERROR_CERT_WAS_REVOKED' => 0x3060,
          'eECAS_ERROR_CERT_WAS_REJECTED' => 0x3061,
          'eECAS_CHALLENGE_MISSMATCH' => 0x3069,
          'eECAS_RENEW_NO_IN_RENEWAL_PERIOD' => 0x3300,
          'eECAS_RENEW_PARENT_CERT_INVALID' => 0x3301,
          'eECAS_RENEW_PARENT_CERT_REVOKED' => 0x3302,
          'eECAS_RENEW_SUBJECT_DN_DOES_NOT_MATCH' => 0x3303,
          'eECAS_REPLACE_PARENT_CERT_INVALID' => 0x3304,
          'eECAS_REPLACE_PARENT_CERT_EXPIRED' => 0x3305,
          'VICE_CERT_ASSOCIATED_WITH_THIS_TRANS_ID_HAS_BEEN_REVOKED' => 0x391f,
          'VICE_CERT_ASSOCIATED_WITH_THIS_TRANS_ID_HAS_BEEN_REJECTED' => 0x3920,
          'VICE_CERT_ASSOCIATED_WITH_THIS_TRANS_ID_HAS_EXPIRED' => 0x3933,
          'VICE_PICKUP_NOT_FOUND' => 0x391b,
          'ERROR_FOUND_IN_LIST' => 0x6400 );

    %VICE_error_meanings =
        ( 0x0000 => 'Success',
          0x0100 => 'System out of memory',
          0x0105 => 'Unknown attribute tag',
          0x0106 => 'Unknown attribute type',
          0x0107 => 'Invalid attribute value',
          0x0108 => 'Attribute value too long',
          0x010a => 'BER encoding error',
          0x010c => "Can't build or verify cert chain",
          0x010d => 'Certificate encoding error',
          0x0110 => 'Certificate Revocation List (CRL) encoding error',
          0x0114 => 'Unsupported data encryption algorithm',
          0x0115 => 'Unknown data encryption algorithm',
          0x0120 => 'I/O error (VICE backend system may be unavailable)',
          0x0125 => 'Unsupported message digest algorithm',
          0x0126 => 'Unknown message digest algorithm',
          0x0128 => 'Invalid domain name (DN)',
          0x012a =>
          'Certificate Revocation List (CRL), certificate or account not found',
          0x0132 => 'Unsupported password-based encryption algorithm',
          0x0133 => 'Unknown password-based encryption algorithm',
          0x013c => 'Unsupported certificate, CRL or message signature algorithm',
          0x013d => 'Unknown certificate, CRL or message signature algorithm',
          0x0145 => 'Invalid authorityKeyID extension',
          0x0146 => 'Invalid keyUsage extension',
          0x0147 => 'Invalid basicConstraint extension',
          0x0148 => 'Invalid privateKeyUsage extension',
          0x014a => 'Invalid subjectAltName extension',
          0x1005 => 'An already valid, non-expired, non-revoked certificate '
          . 'currently exists with the same domain name (DN).',
          0x3041 => 'System configuration error.  Contact ' . $VICE_support,
          0x3047 => 'System configuration error.  Contact ' . $VICE_support,
          0x3060 => 'The certificate in question has been revoked',
          0x3061 => 'The certificate in question has been rejected',
          0x3069 => 'The challenge phrase provided was incorrect',
          0x3300 => 'A message was received attempting to renew a certificate which '
          . 'is not in its renewal period',
          0x3301 => 'A message was received attempting to renew a certificate which '
          . 'is invalid',
          0x3302 => 'A message was received attempting to renew a certificate which '
          . 'was revoked',
          0x3303 => 'A message was received attempting to renew or replace '
          . 'a certificate with a different subject DN from the original',
          0x3304 => 'A message was received attempting to replace a certificate which '
          . 'is invalid',
          0x3305 => 'A message was received attempting to replace a certificate which '
          . 'has expired',
          0x391f => 'An attempt was made to pickup a certificate which was revoked',
          0x3920 => 'An attempt was made to pickup a certificate which was rejected',
          0x3933 => 'An attempt was made to pickup a certificate which has expired',
          0x391b => 'The certificate was not found',
          0x6400 => 'A required field was missing from the reginfo, or '
          . 'one or more fields failed input validation' );

    %VICE_error_names = (); # reverse lookup for %VICE_error_codes.

    my @bad = ();
    while (my ($k,$v) = each %VICE_error_codes ) {
        $VICE_error_names{$v} = $k;
        push @bad, $v unless defined $VICE_error_meanings{$v}
    }
    warn "VICE error codes out of sync with meanings: @bad\n" if @bad;
}

# describe_error(%data)
#
# Returns a string describing the error contained in a VICE query response
# packaged (in the manner of $zv->_query) as a hash.
sub describe_error {
    my %data = @_;
    my ($status, $hcode) = ($data{status}, $data{error_code});
    my $ecode = hex($hcode);
    my $fail = (defined $status) ? "($status)" : '(???)';
    my ($vice, $tail) = ("; contact $VICE_support", '');
    my @inval;

    @inval = split /:/, $data{invalid_input_fields} if $data{invalid_input_fields};
    push @inval, split(/:/, $data{reginfo_error_fields}) if $data{reginfo_error_fields};

    if ( @inval ) {
        $tail = "<BR>\nMissing or invalid data: " . join(', ', @inval);
    }

    if ( defined $ecode ) {
        my $nom = $VICE_error_names{$ecode};
        if ( $nom ) {
            $fail .= " $nom: $VICE_error_meanings{$ecode}";
            $vice = '';
        } else {
            $fail .= " $hcode: undocumented VICE error code";
        }
    } elsif ( defined $hcode ) {
        $fail .= ": non-numeric error code, $hcode";
    } else {
        $fail .= ": no error code specified";
    }

    my $worry = '';
    while ( my ($k,$v) = each %data ) {
        next if $k =~ /^(status|certificate|transaction_id|error_code|invalid_input_fields|reginfo_error_fields)$/o;
        if ( $worry ) { $worry .= ', '; }
        else { $worry = "<BR>\nSurprising data in VICE response:"; }
        $worry .= " $k => $v"
    }

    return "$fail$vice$tail$worry";
}

# show_error_code_table(voice)
#
# &$voice()s an HTML table mimicking the one in Appendix A (Error Code
# Descriptions) of the VICE spec (to enable checking).

sub show_error_code_table {
    my $vox = shift;
    my ($k, $v);
    &$vox("<table><tr><th>Hex code<th>Error Name<th>Error Description</tr>\n");
    foreach $k ( sort keys %VICE_error_meanings ) {
        &$vox(
         "<tr><td>$k<td>$VICE_error_names{$k}<td>$VICE_error_meanings{$k}</tr>\n");
    }
    &$vox("</table>\n");
}

# _fresh_name(stem)
#
# General utility: returns a filename not currently in use by extending stem.
# Extension is just a block of digits.
sub _fresh_name {
    my $stem = shift;
    my $tail = "0";
    $tail++ while -e "$stem$tail";
    return "$stem$tail";
}

use Z; # for quoting and unquoting of URLs and HTML.

# $zv->_query( query_string )
#
# For internal use by Vice.pm only.
#
# POSTs query_string to $zv's vice_url, parses response into a hash.  If $zv has
# a voice, it will be called on message texts to report how the enquiry
# proceeds.  Returns the response hash on success; on failure, it returns a hash
# whose one key is 'Zfailed', with value a string indicating the problem.  On
# success, caller should check the `status' entry in the hash returned: if this
# is not `success' then the query was successfully made and answered, but the
# answer wasn't what you wanted.

sub _query {
    my ( $self, $query ) = @_;
    my ( $url, $vox ) = ( $self->{query}, $self->{speak} );
    return ( 'Zfailed' => 'Bad query specification' ) unless ($query and $url);
    my $auth; # a bodge for the sake of the fake server.
    if ( $url =~ /\?auth=([^&]+)/ ) {
        ($url, $auth) = ($`, $1);
        my $tail = $';
        $url .= "?$'" if $tail =~ /^&/;
    }

    my $httpclient = "$ENV{ZEUSHOME}/admin/bin/httpclient";
    my $htargs;
    if ( $auth ) {
        $httpclient .= ' -method POST';
        $htargs = "-auth $auth $url";
    } else {
        $htargs = "-method POST $url";;
    }

    # Get our httpclient to do the job:
    my $cmd = "echo '$query' | $httpclient $htargs";
    # pretty-print $cmd in output:
    &$vox( "<PRE>echo ...query data... | $httpclient \\\n\t $htargs\n" )
        if $vox;

    my $fail = ( open VICE, "$cmd |" ) ? '' : "Unable to invoke $httpclient: $!";

    # Read to pre-header line:
    unless ( $fail ) {
        while ( <VICE> ) {
            &$vox($_) if $vox;
            if ( m|HTTP/([0-9.]+) ([0-9]+) | ) {
                my ($version, $status, $message) = ($1, $2, $');
                $fail = "Bad HTTP/$version status, $status: $message"
                    unless $status == 200;
                last;
            }
        }
    }
    # Read headers:
    unless ( $fail ) {
        while ( <VICE> ) {
            &$vox($_) if $vox;
            last if /^\s*$/;
        }
    }
    &$vox( "</PRE>\n" ) if $vox;

    # Slurp up the rest:
    my $text;
    unless ( $fail ) {
        local $/;
        # Collect whole of remaining input:
        undef $/; $text = <VICE>;
        # Chomp any trailing newlines:
        $/ = ''; chomp $text;
    }

    if ( $fail ) {
        close VICE;
        return ( 'Zfailed' => $fail );
    }
    close VICE or return ('Zfailed' => "Failed to run $httpclient: $!");
    return ( 'Zfailed' => "Empty response from $url" ) unless $text;

    # Parse this as application/x-www-form-urlencoded data:
    my %data = ();
    foreach (split /&/ms, $text) {
        next unless /=/;
        my ($k, $v) = (Z::unquote_input($`), Z::unquote_input($'));
        if ( $data{$k} ) {
            $data{$k} .= ($/ || "\n") . $v;
        } else {
            $data{$k} = $v;
        }
    }

    # Report on what got slurped:
    if ( $vox and keys %data ) {
        # Report on what we read: first, tart it up a bit.
        &$vox( '<P>Form contained data: </P><DL>', "\n" );
        while ( my ($k, $v) = each %data ) {
            $v = &Z::quote_html($v); # enquote it
            &$vox( "<DT>$k<DD><PRE>$v</PRE></DD>\n" );
        }
        &$vox( "</DL>\n" );
    }

    # and return it.
    return %data;
}

# $v = new Vice([voice, [certificate]])
#
# Stores data on self.  See set_* methods for other (checked/munged) data.
# The value of `voice' should be either false or (a reference to) a
# function to call on text which the user might wish to see.
#
# Various other attributes of $v may be set using the set_* methods below:
# these are separate because each attribute needs to incorporate both config
# data and form data; the latter should override the former, but the config
# should be able to contribute information missing from the form; and these
# attributes are (at least in principle) subject to validity checks.

sub new {
    my $class = shift;
    my $self = {};

    my ($vox, $cert) = @_;
    $self->{speak} = $vox if $vox;
    $self->{certificate} = $cert if $cert;
    # Other attributes may be set other ways:
    # {request}, {pay}, {query}, {id}

    return bless $self, $class;
}

# $v->transid([request])
#
# Asks $v->{query} for a transaction ID with which to handle the certificate
# signing request: stores the new ID in $v->{id}.
# If no request is given, the contents of $v->{request}, if any, are used.
# Returns undef on success, else a string indicating the cause of failure.
sub transid {
    my ($self, $sku, $ver) = @_;
    return 'Prior transaction would be forgotten' if $self->{id};

    return 'No VICE URL in form data, needed for contact with VeriSign'
        unless $self->{query};

    return 'No certificate signing request specified' unless $self->{request};
    my $csr;
    { local $/ = undef; $csr = `cat $self->{request}`; }
    my %data = $self->_query( join '&',
                              "request_type=TransactionID",
                              "sku=$sku",
                              "application=ZeusServer",
                              "application_version=$ver",
                              "csr=" . &Z::quote_url($csr) );
    return $data{Zfailed} if defined $data{Zfailed};
    return &describe_error(%data) unless $data{status} =~ 'success';
    return 'Failed to obtain transaction ID' unless $data{transaction_id};

    $self->{id} = $data{transaction_id};
    return undef;
}

# $v->pickup()
#
# Asks $v->{query} how the transaction is getting along; collects certificate,
# if ready, as $v->{certificate}.
sub pickup {
    my ($self) = @_;
    return 'No pending VICE order' unless ($self->{query} && $self->{id});

    my %data = $self->_query( "request_type=Pickup&transaction_id=$self->{id}" );
    return $data{Zfailed} if defined $data{Zfailed};

    if ( $data{status} =~ 'success' ) {
        $self->{certificate} = $data{certificate};
    } elsif ( $data{status} !~ 'pending' ) {
        return &describe_error(%data);
    }

    return undef;
}

# $zc->save_cert(filename, [overwrite])
#
# Attempts to store $v->{certificate} in the named file.
sub save_cert {
    my $self = shift;
    return 'No collected certificate to save to file' unless $self->{certificate};
    my $fallback;

    if ( $self->{public} ) {
        # This probably can't ever happen.
        $fallback = _fresh_name( "$self->{public}-$$-" );
        rename $self->{public}, $fallback;

    } else { $self->spawn( 'public' ); }
    # assert: $self->{public} but not -f $self->{public}

    my $fail;
    $fail = "Unable to write $self->{public}: $!" unless open CERT, ">$self->{public}";
    print CERT $self->{certificate} unless $fail;
    $fail = "Failed to write $self->{public}: $!" unless ($fail or close CERT);
    if ( $fail ) {
        unlink $self->{public}; # probably a no-op

    } else {# Check new certificate
        use ZSSL;
        my $zs = new ZSSL();
        if ( $self->{request} ) {
            my %h = $zs->readCert( $self->{request} );
            $zs->setDistinguishedName( %{$h{subject}} );
        }

        $fail = $zs->checkCert( $self->{public}, $self->{key} );
        if ( $fail ) {
            my $newname = _fresh_name( "$self->{public}.broken." );
            rename $self->{public}, $newname;
            $fail = "$newname: $fail";
        }
    }

    if ( $fail ) {
        # Restore status quo ante and return.
        rename $fallback, $self->{public} if $fallback;
        # but don't unset $self->{public} - needed in caller's error reports.
        return $fail;
    }

    # success - commit to losing old file (if any), chmod new one.
    unlink $fallback if $fallback;
    chmod 0444, $self->{public};
    return undef;
}

# $v->spawn( part )
#
# With $part =~ /request|public/ this returns the value $v->{$part} would have
# if its file existed already.  Otherwise, acts as a ->get() would, except that
# it sends a warning to STDERR.
sub spawn {
    my ($self, $what) = @_;
    my $stem = "$self->{zdir}/vice";
    if ( $self->{$what} ) {
        print STDERR "\nWarning: internal problems vice-&gt;spawn($what).\n";
    } elsif ( $what eq 'request' ) {
        $self->{request} = $stem . 'CSR';
    } elsif ( $what eq 'public' ) {
        $self->{public} = $stem . 'Cert';
    }
    return $self->{$what};
}

use ZConfig;

# $v->load( servername )
#
# Loads VICE config data for the given server.
sub load {
    my ($self, $site) = @_;
    my $dir = "$ENV{ZEUSHOME}";
    foreach ( 'zip', 'sites', $site ) {
        $dir .= "/$_";
        mkdir $dir, 0755 unless -d $dir;
    }
    $self->{zdir} = $dir; # which exists, by now.
    my $v = "$dir/viceCSR"; # exists from step 4 to successful pickup.
    $self->{request} = $v if -f $v;
    $v = "$dir/viceCert"; # exists after successful pickup
    $self->{public} = $v if -f $v;

    my $zc = new ZConf;
    $zc->load( "$dir/config" ) or return;
    $v = $zc->get( 'vice!private' ); # exists from step 1 onwards
    $self->{key} = $v if $v;
    $v = $zc->get( 'vice!vice_url' ); # exists from step 4 to successful pickup
    $self->{query} = $v if $v;
    $v = $zc->get( 'vice!transaction_id' ); # step 4 onwards
    $self->{id} = $v if $v;
    $v = $zc->get( 'vice!pay_url' ); # step 4 to step 7 (or non-fail pickup)
    $self->{pay} = $v if $v;
    return $dir;
}

# $v->clear();
# Very drastic - clears data and ensures ->save() will *delete* viceCSR and viceCert. 
sub clear {
    my $self = shift;
    foreach ( 'private', 'query', 'id', 'pay', 'request', 'public', 'key' ) {
        delete $self->{$_};
    }
}

# $v->save();
#
sub save {
    my $self = shift;
    my $dir = $self->{zdir};
    return unless $dir; # we never loaded
    unlink "$dir/viceCSR" unless $self->{request};
    unlink "$dir/viceCert" unless $self->{public};

    my $zc = new ZConf;
    $zc->load( "$dir/config" );
    $zc->delete( 'vice' );
    $zc->add( 'vice!private', $self->{key} ) if $self->{key};
    $zc->add( 'vice!vice_url', $self->{query} ) if $self->{query};
    $zc->add( 'vice!transaction_id', $self->{id} ) if $self->{id};
    $zc->add( 'vice!pay_url', $self->{pay} ) if $self->{pay};
    $zc->save();
}

1;
