# ----------------------------------------------------------
# Core CGI-style use routines for AHTML pages.
#

package Z;

use ZConfig;

# Some colours which we might like to configure
# The colour used by Z::warn and Z::error.
# This is configured via Z::set_errorwarn_colour and can be retreived with
# Z::get_errorwarn_colour
$error_foreground_colour="#990000";

# The colour used when replacing '<!--msg: error-->" sections, normally
# the same as the above, but configurable in case you have different coloured backgrounds
# (as in ZAE)
# This is configured via Z::set_errorheading_colour and can be retreived with
# Z::get_errorheading_colour
$error_heading_colour="#990000";

# The error message to use at the top of any page containing errors
# can be configured with Z::set_errorheading.  The default should
# normally be fine, but may not always be applicable

# Z::write string
#
# Write output to the client.


sub write { $main::output .= join("", @_); }


sub set_settings {
   my( %defaults ) = @_;

   # base -> base for config keys
   # configfile -> specify config file location overriding server  

   # Get the user's settings and save them
   import_query_string();

   # make sure that there is only one back value
   if ( defined $::f{"back"} ) {
       ($::f{"back"}) = split /\s/, $::f{"back"};
   }
   my( %f ) = %::f;
   
   # install the defaults
   foreach( keys %defaults ) { $::f{$_} = $defaults{$_}; }

   # read our config settings
   import_config_settings( $defaults{configfile} );
   
   # override with our posted values
   foreach( keys %f ) { 
      # New key/value pairs are added from the form elements %<id>_<prefix>
      # and %<id>!<suffix>.  For example, if $::f{base} == modules, to add
      # modules!mime!types!pdf application/pdf, submit the form elements:
      # base=>modules, %1_mime!types=>pdf, %1=>application/pdf
      
      if( /^\%/ and $f{$_} ) {
         if( /^\%([^!]+?)_(.+)$/ ) {
            ($id, $prefix, $key) = ($1, $2, $f{$_});
            if( defined $f{"\%$id"} ) { $::f{"!$prefix!$key"} = $f{"\%$id"}; }
            
            # Handle child nodes:
            foreach( grep /^\%$id!/, keys %f ) {
               if( /^\%$id!(.*)$/ ) {
                  ($leaf) = ($1);
                  
                  $::f{"!$prefix!$key!$leaf"} = $f{"\%$id!$leaf"};
               }
            }
         }
      } else {
         $::f{$_} = $f{$_}; 
      }
   }
}

# Z::import_query_string
#
# Parse the query string and setup a hash for it
# if a hash of names and defaults are given only those are imported
# and their values are initially set to a default value
# Values are imported into %::f
# If it's a multipart file post form, filenames will appear in $::f variables,
# and the file contents will appear in $::file{filename} variables

sub import_query_string {
   # Is it a multipart POST request?
   my $is_multipart = exists $ENV{CONTENT_TYPE} &&  $ENV{CONTENT_TYPE} =~ m/multipart\/form-data/i;

   if ( defined( $query_string ) ) {
      return;
   }

   $ENV{REQUEST_METHOD} = 'GET' unless defined $ENV{REQUEST_METHOD};
   
   $query_string = "";

   if ( $ENV{REQUEST_METHOD} eq 'POST' ) {
      if ( ! $is_multipart ) {
         # Multiparts are handled separately later
         # Get query_string from STDIN
         # Ignore content length for now
         local $/=undef;
         $query_string = <STDIN>;
      }

      if ( $ENV{QUERY_STRING} ) {
         # get this too 
         $query_string .= '&' . $ENV{QUERY_STRING};
      }
   } elsif ( $ENV{REQUEST_METHOD} =~ /^(GET|HEAD)$/ ) {
      $query_string = $ENV{QUERY_STRING};
   }
      
   # import into %::f
      
   my ( $key, $value );
   foreach $pair ( split /&/ms, $query_string ) {
      next if (!$pair);
      ( $key, $value ) = split /=/ms, $pair, 2;
      
      $key = &unquote_input($key);
      $value = &unquote_input($value);
      
      if( defined $::f{$key} ) {
         $::f{$key} = join $/||"\n", $::f{$key}, $value;
      } else {
         $::f{$key} = $value;
      }
   }

   if ( $is_multipart ) {
      _handle_multipart();
   }
}

sub _handle_multipart
{
    $boundary = $ENV{CONTENT_TYPE};
    $boundary =~ s/^.*?=//;
    $boundary =~ s/^\ *//;
    $boundary =~ s/\ *$//;
    my ($name, $filename, $value);
    $name = ""; $filename = ""; $value = "";
    $state = 0;

    while ( <STDIN> ) {
        $actual = $_;

        s/(\n|\r)$//g;
        
        if ( m/^--$boundary/ ) {
            if ( $name ne "" and $value ne "" ) {
                $value =~ s/(\r\n|\n\r|\n)$//s;
                if ( $filename eq "" ) {
                    $::f{$name} = $value;
                }
                else {
                    $::f{$name} = $filename;
                    $::file{$filename} = $value;
                }
            }
            $filename = "";
            $name = "";
            $value = "";
            $state = 1;
            
            last if ( m/^$boundary--/ );
            next;
        }
        elsif ( $state == 1 ) { 
            if (m/^Content-Disposition/) {
                s/.*?:\ //;
                @content = split /;/, $_;
                
                while (@content){
                    $_ = shift @content;
                    next if ( ! m/=/ );
                    s/\s*$//;
                    s/^\s*//;
                    ($first, $second) = split /=/, $_, 2;
                    $second =~ s/^\"//;
                    $second =~ s/\"$//;
                    if ( $first eq "name") {
                        $name = $second;
                    }
                    if ( $first eq "filename" ) {
                        $filename = $second;
                    }
                }
            }
            elsif ( m/^Content-Type/ ) {
            }
            elsif ( $_ eq "" ) {$state = 2; }
        }
        elsif ( $state == 2 ) {
            $value .= $actual ;
        }
    }
}

sub import_config_settings {
   my ( $configfile ) = @_; # optional
   my( $zc ) = new ZConf;

   if ( defined $configfile ) {
	return if ( $configfile eq "" );
   } else {
      if( defined $::f{server} ) {
         $configfile = "$ENV{ZEUSHOME}/webadmin/conf/sites/$::f{server}";
      } else {
         return;
      }
   } 
   
   # need to get server name from query_string
   $zc->load( $configfile ) || ::Error (<<HERE);
Can\'t read config file $configfile
HERE
           
   $::f{configfile} = $configfile;
   $::f{base} = "" if( ! defined $::f{base} );
   
   foreach $key ( $zc->ZConf::listb( $::f{base} ) ) {
      my( $fqk ) = $::f{base}?"$::f{base}!$key":"$key";
      $::f{"!".$key} = $zc->ZConf::get( $fqk );
   }
}

# Z::f(name)
#
# Access to %::f assoc. array with html escaping

sub f {
   my ( $name ) = @_;
    
   if ( ! defined $::f{$name} ) {
      return "";
   } else {
      return &quote_html( $::f{$name} );
   }
}

# check if a value is defined and not empty ( can be 0 )

sub isset { return( defined($_[0]) && $_[0] ne "" ); }

# Z::hide( list )
#
# Returns html for hidden parameters
#
# Put together the html hidden inputs for the given field names

sub hide {
   my @hidden = @_;
   my $var;
   my $value;
   my $html = "";

   foreach $var ( @hidden ) {
      $value = $::f{$var};
      $value = "" if ! defined $value;
      
      $value = &quote_url( $value );
      
      $html .= "<input type=hidden name=\"$var\" value=\"$value\">\n";
   }
    
   return $html;
}


# Z::is_submit()
#
# Check whether this a first time thing or a submission

sub is_submit {
   return ( defined $::f{issubmit} and $::f{issubmit} eq $::f_name );
}


# Z::authenticate
#
# Print authentification headers and body

sub authenticate {
   print qq{WWW-Authenticate: basic realm=\"$_[0]\"
Status: 401 Authenticate

<html>
<head><title>Authentication required</title></head>
<body bgcolor="#ffffff">
<font size=+5>Authentication required:</font><br>
Security clearance needed for $_[0].
</body></html>
};
   exit 0;
}

# Use the md5 binary in the admin server to hash some data and 
# return a PEM encoded string

sub encrypt {
   my ( $input ) = @_;
   my $output = "";
   
   my $md5 = "$ENV{ZEUSHOME}/admin/bin/md5";
   
   local( *IN_PAR, *IN_CHLD, *OUT_CHLD, *OUT_PAR );
   local $SIG{PIPE} = 'IGNORE'; # let syswrite fail and handle that
   
   # SIGCHLD handler unnecessary as all of our filehandles _will_
   # be closed and the short md5 program _will_ exit properly

   pipe IN_CHLD, IN_PAR;    # pass input to child's STDIN
   pipe OUT_PAR, OUT_CHLD;  # read output from child's STDOUT
  
   ::Error("Executable $md5 not found") if ! -x $md5;

   my( $pid, $outbuf);
   defined( $pid = fork ) or ::Error("Cannot fork: $!");

   if ( $pid != 0 ) {
      # parent
      # close what we don't use
      close OUT_CHLD;
      close IN_CHLD;
      
      # write input to child
      syswrite(IN_PAR, $input, length $input) or ::Error("Writing to md5 failed $!");
      close IN_PAR;
      
      # read child's output
      while ( sysread(OUT_PAR, $outbuf, 128 ) > 0 ) {
         $output .= $outbuf; $outbuf = "";
      }
      close OUT_PAR;

      # wait for the child to exit
      waitpid($pid, 0);
      if ( $? >> 8 ) {
         ::Error("Child md5 returned an error");
      }

   } else {
      # child
      # redirect our ins and outs
      # let the stderr remain on our stderr
      close STDIN;
      open STDIN, "<&IN_CHLD";
      close STDOUT;
      open STDOUT, ">&OUT_CHLD";
     
      # close parent ones we don't need
      close IN_PAR;
      close OUT_PAR; 

      # exec the command
      exec( $md5, '-pem', '-' ) or ::Error("Could not exec md5 $!");
      # error if we get here
      exit( 1 );
   }  
   
   chomp $output;
   return $output;
}


# Z::input_test
#
# Returns true if the external input is acceptable
# false if it is not
# For the purposes of cleaning input destined for backticks etc.

sub input_test {
    my ( $val ) = @_;
    
    if ( $val =~ /[^A-Za-z0-9_\.\-+=]/ ) {
       # Healthy paranoia - if other characters are required then they can be
       # added on the command line
       return undef;
    }
    else {
       return 1;
    }
}


    
# Z::redirect location
#
# Redirects a client to another page, does not return

sub redirect {
   my( $location, $query_string ) = @_;

   # redirect with anchors appears broken on some netscape browsers
   $location =~ s/\#[^\?]*//;
   
   if( $query_string) {
      if( $location =~ /\?/ ) {
         $location .= "\&$query_string";
      } else {
         $location .= "\?$query_string";
      }
   }
   
   if( $location =~ /\Q$ENV{REQUEST_URI}\E$/ ) {
      $location .= ($location =~ /\?/)?"\&cache=no":"?cache=no";
   }

   print qq{Location: $location
Content-Type: text/html
Pragma: no-cache

<html>
<head><title>Moved</title></head>
<body bgcolor="#ffffff">
That page has moved to <a href=\"$location\">$location</a>
</body></html>
};
    exit 0;
}


# Z::parse_url
#
# take a full or fragmented url and set up an array containing
# ( prefix, hostname, port, path ) making sensible guesses
#
# if a second arg is given it should be either 'file' or 'hostname'
# 
# To distinguish usefully between an unqualified file and a hostname
# in an abbreviated url - sensible guesses are taken of course

sub parse_url {
   my ( $url, $type ) = @_;
   my ( $prefix, $hostname, $port, $path, $query_string ) = 
         ( "", "", "", "", "" );
   my $file_mode = 1; # assume anything debatable is a file
   
   if ( defined $type && $type =~ /^h/ ) {
      $file_mode = 0;  # unless otherwise stated
   }
   
   # QUERY STRING
   
   # Get any query string off early as '?' is pretty distinctive
   # If it's not found it's not there
   
   if ( $url =~ s/\?(.*)$// ) {
      $query_string = $1;
   }
   
   # PREFIX
   
   if ( $url !~ s!^(\w{2,6}):\/\/!! ) {
      # Guess 
      if ( defined $ENV{HTTPS} && $ENV{HTTPS} =~ /on/i ) {
         # This is unlikely to be set without good reason
         # Use partial urls and let this find the connection type
         
         $prefix = 'https';
      } else {
         $prefix = 'http';
      }
   } else {
      $prefix = $1;
      $file_mode = 0; # Not a filename
   }
   
   # Hostname/filename final detection
   
   if ( $url !~ m!^/!    # If it doesn't start with a slash
      and (  
         # It has dots or colons before the first slash
         $url =~ m!([^\/]*)/! && $1 =~ m![\.:]! 
         or
         $url =~ m!^\w+(\.\w+){2,}!   # or starts word.word.word....
      ) ) {
      # Looks more like a hostname than a filename
      $file_mode = 0;
   }
   
   # HOSTNAME
   
   # From now, if we're still in file mode take guesses.
   # Everything else goes into the file path
   
   # Grab hostname (.....):8012 or (....)/foo.bar or guess
   
   my $hostmatch;
   
   if ( ! $file_mode ) {
      # if we're looking for hostnames
      if ( $hostmatch = $url =~ s!^([^:/]+)!! ) {
         $hostname = $1;
      }
   }
   
   if ( $file_mode or ! $hostmatch ) {
      # guess
      
      $hostname = $ENV{SERVER_NAME} if defined $ENV{SERVER_NAME};
      $hostname = $ENV{HTTP_HOST} if( defined $ENV{HTTP_HOST} );
      $hostname =~ s/:\d+//; # remove port number
   }
   
   # PORT
   
   # Grab port if there is one or guess
   
   my $port_match;
   
   if ( ! $file_mode ) {
      # if we're looking for a full url
      if ( $port_match = $url =~ s!^:(\d+)!! ) {
         $port = $1;
      }
   }
   
   if ( $file_mode or ! $port_match ) {
      # guess
      
      if ( defined $ENV{SERVER_PORT} ) {
         $port = $ENV{SERVER_PORT};
      } else {
         if ( $prefix eq 'https' ) {
            $port = 443;
         } else {
            $port = 80;
         }
      }
   }
   
   # PATH
   
   # Assume that the remaining is a file path
   
   if ( $url =~ m!^\/! ) {
      # absolute relative to docroot
      $path = $url;
   } else {
      if ( $url ) {
         # relative to PATH_INFO
         if ( defined $ENV{REQUEST_URI} ) {
            $path = $ENV{REQUEST_URI};
            $path =~ s!\/[^\/]*$!!; # remove any trailing filename and slash
         }
         $path .= '/'.$url;
      }   
   }
   
   return ( $prefix, $hostname, $port, $path, $query_string );
}


# Z::back
#
# Redirects the user back to the previous page, or "/" if there
# is no referer.

sub back {
   if ( defined $ENV{HTTP_REFERER} ) { redirect $ENV{HTTP_REFERER}; }
   else                              { redirect "/"; }
}


# Z::unquote_input
#
# Unescape URL-encoded data

sub unquote_input {
   my($todecode) = @_;
   if (! defined($todecode)) {
     return undef;
   }
   $todecode =~ tr/+/ /;       # pluses become spaces
   $todecode =~ s/%([0-9a-fA-F]{2})/pack("c",hex($1))/ge;
   return $todecode;
}


# Z::quote_url
#
# return string quoted appropriately to appear in a url

sub quote_url {
   my($toencode) = @_;
   $toencode=~s/([^a-zA-Z0-9_\-.])/uc sprintf("%%%02x",ord($1))/eg;
   return $toencode;
}


#
# Z::PEMencode
#
# return string PEM encoded.  PEM is a 6bit printable encoding.

sub PEMencode
{
   my @pem_enc_table =
     (
      65, 66, 67, 68, 69, 70, 71, 72, 73, 74,
      75, 76, 77, 78, 79, 80, 81, 82, 83, 84,
      85, 86, 87, 88, 89, 90, 97, 98, 99,100,
      101,102,103,104,105,106,107,108,109,110,
      111,112,113,114,115,116,117,118,119,120,
      121,122, 48, 49, 50, 51, 52, 53, 54, 55,
      56, 57, 43, 47
     );
   
   my( $p ) = @_;  # string to encode
   my( $r ) = "";  # result
   my( $bitcount ) = 0;
   my( $buffer ) = 0;
   my( $highbits );

   foreach my $c ( split //, $p ) {
      # read in 8 bytes
      $buffer <<= 8;
      $buffer += ord( $c );
      $bitcount += 8;
      while( $bitcount > 5 ) {
         # get high 6 bits
         $highbits  = ( $buffer   >> ($bitcount-6));
         $buffer   -= ( $highbits << ($bitcount-6));
         $bitcount -= 6;
         $r .= chr( $pem_enc_table[$highbits] );
      }
   }
   if( $bitcount ) {
      # could be some remaining bits
      $buffer <<= 6 - $bitcount;
      $r .= chr( $pem_enc_table[$buffer] );
   }
   # pad out to 24bits (4 bytes in base 64)
   my $short_count = length( $r ) & 3;
   if( $short_count ) { $short_count = 4 - $short_count; }
   while( $short_count-- ) { $r .= "="; }

   return $r;
}


#
# Z::PEMdecode
#
# Given a PEM encoded string, returns the decoded version.

sub PEMdecode
{
   my @pem_dec_table = 
     (
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,62,-1,-1,-1,63,52,53,54,55,56,57,58,59,60,61,-1,-1,
      -1,-1,-1,-1,-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,
      15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1,-1,26,27,28,
      29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,
      49,50,51,-1,-1,-1,-1,-1
     );
   
   my( $p ) = @_; # string to decode
   my( $r ) = ""; # result
   my( $bitcount ) = 0;
   my( $buffer ) = 0;

   foreach my $c ( split //, $p ) {
      $c = ord( $c );
      next if( $c > 127 || $pem_dec_table[$c] < 0 );
      $buffer <<=6;
      $buffer   += $pem_dec_table[$c];
      $bitcount += 6;
      if( $bitcount > 7 ) {
         $bitcount -= 8;
         $r .= chr( $buffer >> $bitcount );
      }
   }
   return $r;
}



# Z::quote_html
#
# return string but with html-special characters escaped,
# necessary if you want to send unknown text to an html-formatted page.

sub quote_html {
   my($toencode) = @_;
   if (! defined($toencode)) {
     return undef;
   }
   $toencode =~ s/&/\&amp;/g;
   $toencode =~ s/\"/&quot;/g;
   $toencode =~ s/</&lt;/g;
   $toencode =~ s/>/&gt;/g;
   return $toencode;
}


# Z::unquote_html
#
# return string but with html-special escaped returned to normal,
# necessary if you want to get quotedhtml back from a page
sub unquote_html {
   my($toencode) = @_;
   if (! defined($toencode)) {
     return undef;
   }
   $toencode =~ s/\&amp;/&/g;
   $toencode =~ s/&quot;/\"/g;
   $toencode =~ s/&lt;/</g;
   $toencode =~ s/&gt;/>/g;
   return $toencode;
}

    

# Z::addheader
#
# Adds the given MIME-header and value to the output, assuming the rest of
# the processing goes normally.
# Expects 2 arguments, the mime-header, and the value

sub addheader {
   my( $key, $value ) = @_;
   $main::headers{$key} = $value;
}


# Z::include
#
# includes a given file at the current position (without parsing it)

sub include {
   my( $filename ) = @_;
   local $/;
   local *F;
   unless( open( F, "$filename" )) {
      Z::write( "<xmp>Unable to open file: $filename: $!</xmp>" );
      return;
   }
   my $data = scalar <F> || "";
   close F;
   Z::write( $data );
}


# Z::parse
#
# includes a given file at the current position, parsing it for Z
# comments (variables will be available to it, and any side effects it
# has will be visible after the parse).

@internal_parse_pwdstack = (); # private to parse!!

sub parse {
   my( $filename ) = @_;
   local *F;

   my $pwd;
   {
       local $/ ="\n";
       ($pwd) = `pwd`; chomp $pwd;
   }

   unless( open( F, "$filename" )) {
     Z::write( "<xmp>Unable to open file: $pwd/$filename: $!</xmp>" );
       return;
   }

   my $data;
   {
      local $/;
      $data = scalar <F> || "";
   }
   close F;

   # recursion is your friend

   # we push the current directory and chdir to the directory containing the
   # code we are parsing, so code being parsed has the invariant that
   # "pwd == directory containg currently parsed code".

   push @internal_parse_pwdstack, $pwd;

   # create directory owning file
   my $dirname = $filename;
   $dirname =~ s/[^\/]*$//;
   chdir( $dirname ) if ( $dirname ne "" );

   main::parse( $data );

   chdir( pop @internal_parse_pwdstack );
}


# Z::warn/error
#
# evaluates $_[0]; outputs $_[1] and inc error count if fails
# note: validate( undef, "foo" ) == validate( "foo" ), hence the unshift

sub warn {
   $_[1] || unshift @_, 0;
    
   return " " if( !is_submit() );
   return ""  if( $_[0] );

   $::f_warn++;
   return "<a name=\"errors\"><font color=\"$error_foreground_colour\">$_[1]</font></a>";
}

sub error {
   $_[1] || unshift @_, 0;
    
   return " " if( !is_submit() );
   return ""  if( $_[0] );

   $::f_errs++;
   return "<a name=\"errors\"><font color=\"$error_foreground_colour\">$_[1]</font></a>";
}

sub set_errorwarn_colour
{
   $error_foreground_colour = $_[0];
}
sub set_errorheading_colour
{
   $error_heading_colour = $_[0];
}
sub set_errorheading
{
   $error_heading = $_[0];
}
sub get_errorwarn_colour
{
   return $errorwarn_colour;
}
sub get_errorheading_colour
{
   return $error_heading_colour;
}

# Z::is_*
#
# Useful verification functions
# Can be used in Z::warn etc. 
sub is_int {
   return ($_[0] =~ /^[0-9]+$/);
}

# start tag, so that a complete block of text can be 
# later removed
sub startTag($) {
    my $tag = shift;
    if ( ! defined $tag ) { return; }
    return "<!-- TAG: $tag -->" ;
}

# Add an end tag, this should match a previously added
# tag, although no check is made
sub endTag($) {
    my $tag = shift;
    if ( ! defined $tag ) { return; }
    return "<!-- ENDTAG: $tag -->" ;
}

# removes the text between the start tag and the
# end tag, should match a tag which has been
# started and ended, although no check is made
sub removeTag($) 
{
    my $tag = shift;
    if ( ! defined $tag ) { return; }

    $main::output =~ 
	s@<!-- TAG: $tag -->.*<!-- ENDTAG: $tag -->@@s;
}


# Z::form
#
# Starts form; creates unique $::f_name id used to track which form we
# submit/return to

sub form {
   my( %args ) = @_;
   $::f_name = $args{name}||++$::f_count;
   
   my %uri = ();

   ($uri{prefix}, $uri{hostname}, 
    $uri{port},   $uri{path}, 
    $uri{query_string} ) = parse_url( $ENV{REQUEST_URI} );   

   my $key;
   foreach $key ( keys %uri ) {
       if ( defined $::f{$key} ) {
	   $uri{$key} = $::f{$key};
       }
   }

   my( $location ) = "$uri{prefix}://$uri{hostname}";
   $location .= ":$uri{port}" if $uri{port};

   # This allows callers to override where we post the
   # form to by specifying their own 'action' variable.
   # We cope with full URLs, slash-prefixed names for this
   # server, and relative names
   if ( defined $args{action} && $args{action} ) {
      if ( $args{action} =~ /^\w{2,6}:/ ) {
         # It's a full URL
         $location = $args{action};
      } else {
         if ( $args{action} =~ m@^/@ ) {
            # It's a full path
            $location .= $args{action};
         } else {
            # It's relative. Get the path directory from the URI
            ( $dir ) = ( $uri{path} =~ m@^(.*/)[^/]*$@ );
            $location .= $dir . $args{action};
         }
      }
   } else {
      $location .= "$uri{path}";
   }

   $target = $args{target}||$location;

   $location .= '?'."$uri{query_string}" if $uri{query_string};

   $back = $::f{"back"}||$ENV{REFERRER}||$target;
 
   # If base is set, us that, else use a empty string
   my( $base ) = defined( $::f{base} ) ? $::f{base} : "";

   # if an enctype was passed, print it out as well
   $enctype = ($args{enctype})?"enctype=\"$args{enctype}\"":"";

   return <<HERE;
<form name="$::f_name" $enctype action="$location" method=POST>
<a name="$::f_name">
<input type=hidden name=issubmit value="$::f_name">
<input type=hidden name=base value="$base">
<input type=hidden name=target value="$target">
<input type=hidden name=back value="$back">
</a>
HERE
}


# Z::msg
#
# To write messages into arbitrary places forwards and backwards in
# the returned HTML.  The endform used this to write warnings/error
# messages

sub msg_write {
   my( $tag, $str) = @_;
   
   $main::output =~ s/<!--msg: $tag-->/$str/;
}


# Z::endform
#
# Closes a form and immediately attempts to commit the changes

sub endform {
   if( is_submit() ) {      
      if( $::f_errs ) {
 	 if( !defined $error_heading || $error_heading eq "" ) {
         # We have to check this here to give $error_heading_colour all the time
         # possible to be set
           $error_heading =
             "<p><b><font color=\"$error_heading_colour\"> " .
             "Your changes have not been applied, because there were errors.<br> " .
             "<a href=\"#errors\">Please correct them</a> and resubmit your configuration. " .
             "</font></b></p>";
	 }
         msg_write( "error", $error_heading );
         return "</form>";
      }

      # commit configuration
      my( $zc ) = new ZConf;

      $zc->ZConf::load( $::f{configfile} );
      #################################################################
      # VALIDATION CHECKING

      my @errlist=();
      my %tablevalidate;
      my %listvalidate;

      if ( defined $::f{validatefile}) {
         $validfile = $::f{validatefile};
      } else {
         $validfile = "$ENV{ZEUSHOME}/webadmin/conf/validate";
      }

      # Load up regexs that the user must match in their input
      if (open(VALID, "< $validfile")) {
         while (<VALID>) {
            next if /^#/;
	    next unless /(.*?)\t+(.*?)\t+(.*)/;
            my ($key, $regex, $txtdesc) = ( $1, $2, $3 );
	    
            # If key starts with a tilde, put it in a list
            # of regexes to check, else put it in a hash table
            if ($key !~ s/^~//) {
               $tablevalidate{$key} = $regex;
               $tabledesc{$key} = $txtdesc;
            } else {
               $listvalidate{$key} = $regex;
               $listdesc{$key} = $txtdesc;
            }
         }
      }

      # Loop through all keys supplied by the form (in hash %::f);
      # define $keyf as non pling prefixed keys, and skip whitespace keys
      foreach $key (grep /^\!/, sort keys %::f) {
         ($keyf = $::f{base}.$key) =~ s/^\!//;
	 next if ($::f{$key} =~ m/^\s*$/);

         # If the values have changed from the original configuration
         # test to see if they match our validation tests
	 if ( ( ! defined($zc->ZConf::get($keyf))    ) ||
	      ( $zc->ZConf::get($keyf) ne $::f{$key} )
	    ) {
            $regex = $tablevalidate{$keyf};
            if ( defined($regex) ) {
               # Check regex from hash against user input
               if ( $::f{$key} !~ m/$regex/ ) {
                  push @errlist, "Your value '$::f{$key}' for $tabledesc{$keyf} isn't valid";
               }
            } else {
               # Check regex from list against user input
               $regex = getkey($keyf, \%listvalidate);
               if ( defined( $regex ) && $::f{$key} !~ m/$regex/ ) {
                  $txtdesc = getkey($keyf, \%listdesc);
                  push @errlist, "Your value '$::f{$key}' for $txtdesc isn't valid";
               }
            }
         }
      }

      # Build a list of errors to report back to the user
      # that they should fix and resubmit
      #
      if ( @errlist > 0 ) {
         foreach $error (@errlist) {
            $list .= "<li>" . $error . "<br>\n";
         }
         msg_write( "error", <<HERE );
<p><b><font color="$error_heading_colour">
Your changes have not been applied, because there were errors:<br>
<ul>
$list
</ul>
Please correct them and resubmit your configuration.
</font></b></p>
HERE
         return "</form>";
      }

      # END VALIDATION CHECKING
      #################################################################

# commit updates/new entries

      foreach $key (grep /^\!/, sort keys %::f) {
         ($keyf = $::f{base}.$key) =~ s/^\!//;
         $zc->ZConf::add($keyf, $::f{$key});
      }

# do deletions now.  Whatever happens, make sure that we're redirected
# so that the deletions are reflected in the display.  This means that
# warnings won't be shown

      my $deleted = 0;
      foreach $key( grep /^\-/, keys %::f ) {
         ($keyf = $::f{base}.$key) =~ s/\-//;
         $keyf =~ s/^!//;
         $zc->ZConf::delete( $keyf );
         $deleted = 1;
      }

      if ( defined $::f{"configfile"} && $::f{"configfile"} ne "" ) {
	$zc->ZConf::save();
      }

      if( !$deleted and $::f_warn ) {
         msg_write( "error", <<HERE );
<p><b><font color="$error_heading_colour">
Your changes have been applied, although there were warnings reported.<br>
<a href="#errors">Please review these warnings</a>.
</font></b></p>
HERE
         return "</form>";
      }

      if ( $::f{target} ) {
         if ( ! defined $::f{server} ) {
            redirect( $::f{target} )
         } else {
            if ( $::f{target} =~ /server=/ ) {
               $::f{target} =~ s/server=[^&]*(&?)/server=$::f{server}$1/;
               redirect( $::f{target})
            } else {
               redirect( $::f{target}, "server=$::f{server}" );
            }
         }
      }
   }
   
   return "</form>";
}

# Z::form_table
#
# Takes args "base", "key" and "value", and enumerates all key/value pairs
# under $f::{base}!base for editing or deletion

sub form_table {
   my( %args ) = @_;
   my( $base ) = join "!", $::f{base}, $args{base};
   my( $html ) = "";
   
   $html = "<table border=2 cellspacing=2 cellpadding=2>\n";

   # Table heading line
   
   $html .= <<HERE;
<tr>
   <td bgcolor="#aaaaaa"><b>$args{key}</b></td>
HERE
   # optional leaf data
   if( defined $args{leaf} ) {
      foreach ( @{$args{leaf}} ) {
         /(.*?) (.*) ([0-9]+)/;
         $html .= <<HERE;
   <td bgcolor="#aaaaaa"><b>$2</b></td>
HERE
      }
   }
   if (defined $args{value}) {
      $html .= <<HERE;
   <td bgcolor="#aaaaaa"><b>$args{value}</b></td>
HERE
   }
   $html .= <<HERE;
   <td>&nbsp;</td>
</tr>
HERE
   
   # Get the currently defined keys
   @keys = grep /^\!$args{base}!/, keys %::f;
   @keys = map { /^\!$args{base}\!([^!]+)/; $1 } @keys;
   %seen = (); @keys = sort { lc($a) cmp lc($b) } grep { ! $seen{$_}++ } @keys;
   
   foreach $key ( @keys ) {
      ($checked) = (defined $::f{"-!$args{base}!$key"})?" checked":"";
      $html .= <<HERE;
<tr>
   <td bgcolor="#cccccc">$key</td>
HERE
      if( defined $args{leaf} ) {
         foreach ( @{$args{leaf}} ) {
            /(.*?) (.*) ([0-9]+)/;
            $keyname = "!$args{base}!$key!$1";
            $value = Z::f($keyname);
         $html .= <<HERE;
   <td bgcolor="#cccccc">
      <input type=text name="$keyname" value="$value" size=$3>
   </td>
HERE
         }
      } 
      if (defined $args{value}) {
         $keyname = "!$args{base}!$key";
         $value = Z::f($keyname);
         $html .= <<HERE;
   <td bgcolor="#cccccc">
      <input type=text name="$keyname" value="$value" size=30>
   </td>
HERE
      }
      $html .= <<HERE;
   <td>
      <input type=checkbox name="-!$args{base}!$key" value="1"$checked> Delete
   </td>
</tr>
HERE
   }

   $html .= <<HERE;
</table>
HERE
   return $html;
   
}

# Return a matching value matching 'match'
sub getkey {
  my ($match, $hash) = @_;

  foreach $key (keys %{$hash}) {
     return $hash->{$key} if ($match =~ m/^$key/);
  }

  return undef;
}

# ----------------------------------------------------------
# This sub will fork, set*id, exec the given command,
# grab it's output and return ($exitcode, $stdout, $stderr).
# If command is an array reference ( i.e. [ "ls", "-l" ] )
# then execvp is used - avoiding some shell parameter passing issues
# Otherwise a normal Perl exec occurs using a shell

sub run_external {
   my( $command, $stdin, $uid, $gid ) = @_;
   my( $exit, $stdout, $stderr );

   $stdin = "" if ! defined $stdin;
   $stdout = $stderr = "";
   
   # localise our filehandles
   
   local( *CHLD_IN, *CHLD_OUT, $CHLD_ERR );
   local( *PRNT_IN, *PRNT_OUT, $PRNT_ERR );

   local $SIG{PIPE} = 'IGNORE';
   
   # from the child's point of view
   pipe CHLD_IN, PRNT_IN;
   pipe PRNT_OUT, CHLD_OUT;
   pipe PRNT_ERR, CHLD_ERR;
   
   my $pid;   
   if( $pid = fork() ) {
      # parent
      
      # close unused filehandles

      close CHLD_IN;
      close CHLD_OUT;
      close CHLD_ERR;
      
      # make some bit vectors of the outputs

      my( $fh_out, $fh_err, $fh_in ) = ( '', '', '' );
      
      my( $fileno_in ) = fileno(PRNT_IN);
      my( $fileno_out ) = fileno(PRNT_OUT);
      my( $fileno_err ) = fileno(PRNT_ERR);

      vec( $fh_in,  $fileno_in,  1 ) = 1;
      vec( $fh_out, $fileno_out, 1 ) = 1;
      vec( $fh_err, $fileno_err, 1 ) = 1;

      my $readset = $fh_out | $fh_err;
      my $writeset = $fh_in;
      my $fullset = $fh_out | $fh_err | $fh_in;

      # whilst there is input, fetch it
      # "unpack( "%b*", $fdset )" should evaluate to the number of bits set in $fdset
      # i.e. 0 when we are no longer interested in any file descriptors
      # If none is waiting this should block until more data arrives or the other end is closed
      
      my( $r, $w, $e );
      while( unpack( "%b*", $fullset ) && ( select( $r=$readset, $w=$writeset, $e=$fullset, undef ))) {
         my $inbuf;
         
         # check for errors
         if( vec( $fullset, $fileno_in, 1 ) && vec( $e, $fileno_in, 1 ) ) {
            vec( $writeset, $fileno_in, 1 ) = 0; # set stdin fd bit to 0
            vec( $fullset, $fileno_in, 1 )  = 0; # set stdin fd bit to 0
            close PRNT_IN;

         # write if possible
         } elsif( vec( $writeset, $fileno_in, 1 ) && vec( $w, $fileno_in, 1 ) ) {
            my $written = syswrite( PRNT_IN, $stdin, 4096 );            

            substr( $stdin, 0, $written ) = "" if defined $written;
            
            # if nothing left to write, remove this from the select masks and
            # close the fh
            if( ! defined $written || length($stdin) < 1 ) {               
               vec( $writeset, $fileno_in, 1 ) = 0; # set stdin fd bit to 0
               vec( $fullset, $fileno_in, 1 )  = 0; # set stdin fd bit to 0               
               close PRNT_IN;
            }
         }

         # read any data waiting on stderr
         if( vec( $readset, $fileno_err, 1 ) && vec( $r, $fileno_err, 1 ) ) {
            my $ret = sysread( PRNT_ERR, $inbuf, 4096, 0 );
            
            # if nothing left to read, remove from select masks and close fh
	    if( ! defined $ret || $ret < 1 ) {
               vec( $readset, $fileno_err, 1 ) = 0; # set stderr fd bit to 0
               vec( $fullset, $fileno_err, 1 ) = 0; # set stderr fd bit to 0 
               close PRNT_ERR;
            } else {              
               $stderr .= $inbuf;
            }
         }

         # check for errors on stderr and if so remove from select masks and close fh
         if( vec( $fullset, $fileno_err, 1 ) && vec( $e, $fileno_err, 1 ) ) {
            vec( $readset, $fileno_err, 1 ) = 0; # set stderr fd bit to 0
            vec( $fullset, $fileno_err, 1 ) = 0; # set stderr fd bit to 0
            close PRNT_ERR;
         }
         
         
         # read any data waiting on stdout
         if( vec( $readset, $fileno_out, 1 ) && vec( $r, $fileno_out, 1 ) ) {
            my $ret = sysread( PRNT_OUT, $inbuf, 4096, 0 );

	    if( ! defined $ret || $ret < 1 ) {
               vec( $readset, $fileno_out, 1 ) = 0; # set stdout fd bit to 0
               vec( $fullset, $fileno_out, 1 ) = 0; # set stdout fd bit to 0
               close PRNT_OUT;
            } else {              
               $stdout .= $inbuf;
            }
         }
            
         # check for errors on stdout and if so remove from select masks and close fh         
         if( vec( $fullset, $fileno_out, 1 ) && vec( $e, $fileno_out, 1 ) ) {
            vec( $readset, $fileno_out, 1 ) = 0; # set stdout fd bit to 0
            vec( $fullset, $fileno_out, 1 ) = 0; # set stdout fd bit to 0
            close PRNT_OUT;
         }
      }

      waitpid $pid, 0;
      
      $exit = $? >> 8;

      return wantarray ? ( $exit, $stdout, $stderr ) : $exit;

   } elsif( defined $pid ) {
      
      # child
      
      # close unneeded fds

      close PRNT_IN;
      close PRNT_OUT;
      close PRNT_ERR;

      # set our uid, euid, gid, egid

      if( defined $gid ) {
         if( $) !~ /^$gid\b/ ) { eval { $) = "$gid $gid" }; }
         if( $) !~ /^$gid\b/ ) {
            ::Error( "Could not set EGID to $gid. $@" );
            return undef;
         }

         if( $( !~ /^$gid\b/ ) { eval { $( = $gid }; }
         if( $( !~ /^$gid\b/ ) {
            ::Error( "Could not set GID to $gid. $@" );
            return undef;
         }
      }

      if( defined $uid ) {         
         if( $< != $uid ) { eval { $< =  $uid }; }
         if( $< != $uid ) {
            ::Error( "Could not set UID to $uid. $@" );
         }

         if( $> != $uid ) { eval { $> =  $uid }; }
         if( $> != $uid ) {
            ::Error( "Could not set EUID to $uid. $@" );
         }
      }
      
      open( STDIN, "<&CHLD_IN" )   or ::Error( "Cannot redirect STDIN. $!" );
      open( STDOUT, ">&CHLD_OUT" ) or ::Error( "Cannot redirect STDOUT. $!" );
      open( STDERR, ">&CHLD_ERR" ) or ::Error( "Cannot redirect STDERR. $!" );

      # close the now dup'ed filehandles

      close CHLD_IN;
      close CHLD_OUT;
      close CHLD_ERR;

      # if this is a reference (to an array), use parameterised exec
      # otherwise exec through a shell
      
      if( ref $command ) {
         $command_txt = join( " ", @$command );
         exec { $command->[0] } @$command;
      } else {
         $command_txt = $command;
         exec $command;
      }

      ::Error( "Exec of '$command_txt' failed $!" );

   } else {
      # error
      ::Error( "Fork failed $!" );
   }

}

# Z::verify_output_file(TYPE, PATH)
#
# This subroutine is used to check that we are able to write some file.
# Various checks are performed, and (hopefuly) informative error messages
# are reported back using ::Error, which should be defined in your
# script.
#
# This is NOT an excuse not to check for error opening the file!  Rather
# it's a pre-open check to catch common problems.  The open might well
# still fail for some reason unrelated to the checks performed here.
#
# A particular oddity is that this subroutine does NOT check whether
# the file is writable.  DO NOT MAKE IT DO SO.  The reason is that the
# first caller for this subroutine attempts to overwrite existing
# files if and only if requested to do so, regardless of permissions,
# and it is desirable that this behaviour be maintained.
#
# TYPE is some kind of description of what the file is; it will appear
#      in error messages.  An example might be "Public key file".
#
# PATH is a path name to check, which is presumably user input.

sub verify_output_file($$) {
   my $type = shift;
   my $path = shift;

   # insist on absolute paths
   # (this rejects empty filenames too)
   if( $path !~ /^\// ) {
      ::Error( "$type '$path' is not an absolute path name; it must begin with a '/'." );
   }

   # reject trailing slashes
   if( $path =~ /\/$/) {
      ::Error( "$type '$path' is not a valid filename; it must not end in '/'." );
   }

   # strip the basename off, so we just have the directory
   my $dir = $path;
   $dir =~ s/\/[^\/]*$//;

   # sanity checks on the directory
   if( -l $dir && ! -e $dir) {
      ::Error( "$type '$path' invalid: '$dir' is a dangling symbolic link" );
   }
   if( ! -e $dir ) {
      ::Error( "$type '$path' invalid: '$dir' does not exist" );
   }
   if( ! -d $dir ) {
      ::Error( "$type '$path' invalid: '$dir' is not a directory" );
   }
   if( ! -w $dir ) {
      ::Error( "$type '$path' invalid: '$dir' is not writable" );
   }

}

1;
