package ZConf;

use ZeusOS;

# $zc = new ZConf([$filename])       # Create new ZConf object

# $zc->load([$filename])      # read in a file

# In memory functions after load

# $value = $zc->get( $key )   # Gets a value given a full key
# $zc->add( $key, $value )    # Inserts or changes a key-value
# @list = $zc->list($prefix)  # gets a list of the next level keys
#                                       matching the given prefix
# $zc->exists( $key )         # Test whether the given key/branch exists
# $zc->delete( $key )         # remove a key branch or value

# $zc->save([$filename])      # write out the file

# undef is returned on failure
# See $! for system errors

# Internally
# $self->{filename}  # source filename of current file
# $self->{file}      # loaded file contents

# Fnctl flags

$O_RDONLY = $ZeusOS::O_RDONLY;
$O_WRONLY = $ZeusOS::O_WRONLY;
$O_CREAT = $ZeusOS::O_CREAT;
$O_TRUNC = $ZeusOS::O_TRUNC;


##################################################################
#
# $zc = new ZConf([$filename]);
#
##################################################################

sub new {
   my $class = shift;
   my $self = {};

   $self->{filename} = shift if @_;

   bless $self, $class;
}

##################################################################
#
# $zc->load([$filename]) : string
#
# returns contents of a file or undef
#
##################################################################

sub load {
   my $self = shift;

   my $filename;
   if ( @_ ) {
      $filename = shift @_;
      $self->{filename} = $filename;
   } else {
      $filename = $self->{filename};
   }

   local *FILE;

   # read in and return file or undef

   # Check that the file is a plain text file
   unless ( -f $filename and -T _ ) {
       $self->{file} = ''; # so we *shall* be able to ->add() !
       return undef;
   }

   sysopen( FILE, $filename, $O_RDONLY );# or return undef;
   local $/ = undef;
   $self->{file} = <FILE>;
   close( FILE ) or return undef;

   return $self->{file};
}


##################################################################
#
# $zc->save([$filename])
#
# Returns the error given if this fails
#
# No locking just a temporary file and moves
#
##################################################################

sub save {
   my $self = shift;

   my $filename;
   if ( @_ ) {
      $filename = shift;
   } else {
      $filename = $self->{filename};
   }

   my $temporary_filename = &temporary_filename( $filename );

   local *FILE;
   local $/ = undef;

   # write out the contents of the temp file
   sysopen( FILE, $temporary_filename, $O_WRONLY|$O_TRUNC|$O_CREAT, 0666 ) or
      return undef;
   print FILE $self->{file} or return undef;
   close FILE or return undef;

   unless ( rename $temporary_filename, $filename ) {
      # unlink temp file
      unlink $temporary_filename;
      return undef;
   }

   return 1;
}

##################################################################
#
# $zc->flatten()
#
# returns a flattened 
#
##################################################################

sub flatten {
   my $self = shift;
   
   return defined $self->{file} ? $self->{file} : "";
}



##################################################################
#
# $value = $zc->get($key) : string
#
# Looks in the config file for the given virtual server and gets the
# value corresponding to the full key given.
# This may be "" if nothing matches or the value is null.
# To differentiate between this case use exists
#
##################################################################

sub get {
   my ( $self, $key ) = @_;
   my $value = "";

   return undef unless defined $self->{file};

   my @file = split /\n/, $self->{file};
   local $_;

LINE:
   foreach $_ ( @file ) {
      next LINE if /^#/; # comments start with ^#

      if ( ( $value ) = /^\Q$key\E[ \t]*(.*)/ )    # value can be ""
      {
         last LINE;     # doesn't cater for multiple matching keys
      }
   }

   return $value;    # "" or value
}



##################################################################
#
# $zc->add( key, value )
#
# set a key value pair in memory
#
##################################################################

sub add {
   my ( $self, $key, $value ) = @_;

   return undef unless defined $self->{file};

   # Fix problems where input validation fails to spot spaces in keys
   $key =~ s/[\s\n]//msg;

   if ( $value =~ /[\n]/ms ) {
      # Having these characters in our value will break things
      # The function setting this field should cater for escaping them
      # if necessary  --  pass an error back by some mechanism
      return undef;
   }

   if ( $key =~ /^#/ ) {
      # Comments are hash at the start of a line so
      # you can't have one in a key
      # pass an error back
      return undef;
   }

   # replace line

   my $line = "$key $value";

   unless ( $self->{file} =~ s/^\Q$key\E[ \t]*.*/$line/m ) {
      # or append it

      $self->{file} =~ s/\n+$/\n/s; # try and keep the end of the file neat
      $self->{file} =~ s/^\n+//s;   # likewise the start
      $self->{file} .= "$line\n";
   }

   # this shouldn't have messed up anything in the file like comments
   return 1;
}



##################################################################
#
# @list = $zc->list( $prefix ) : list of next key level
#
# match the leaf nodes matching the given prefix
# these can then be used to look up values
#
##################################################################

sub list {
   my ( $self, $prefix ) = @_;

   return undef unless defined $self->{file};

   my @file = split /\n/, $self->{file};

   my %keys = ();
   my $key;

   local $_;

LINE:
   foreach $_ ( @file ) {
      next LINE if /^#/; # comments start with ^#
      ( $key ) = $prefix?/^\Q$prefix\E\!([^!\s]+)\s*/:/^([^!\s]+)\s/;
      $keys{$key} = 1 if ( defined $key and $key ne "" );
   }

   my @keys_unsorted = keys %keys;
   my @keys;

   if ( grep /^\D/, @keys_unsorted ) {
      # sort lexically if not all of the keys start with a number
      @keys = sort @keys_unsorted;
   } else {
      # otherwise sort numerically
      @keys = sort { $a <=> $b } @keys_unsorted;
   }

   return @keys;
}


##################################################################
#
# @list = $zc->listb( $prefix ) : list whole branch of keys
#
# match the branch  matching the given prefix
# these can then be used to look up values
#
##################################################################

sub listb {
   my ( $self, $prefix ) = @_;

   return undef unless defined $self->{file};

   my @file = split /\n/, $self->{file};

   my %keys = ();
   my $key;

   local $_;

LINE:
   foreach $_ ( @file ) {
      next LINE if /^#/; # comments start with ^#
      ( $key ) = $prefix ? /^\Q$prefix\E\!(\S+)\s*/ : /^(\S+)/; # Previously : /^(\S+)\s/
      $keys{$key} = 1 if $key;
   }

   my @keys_unsorted = keys %keys;
   my @keys;

   if ( grep /^\D/, @keys_unsorted ) {
      # sort lexically if not all of the keys start with a number
      @keys = sort @keys_unsorted;
   } else {
      # otherwise sort numerically
      @keys = sort { $a <=> $b } @keys_unsorted;
   }

   return @keys;
}



##################################################################
#
# $zc->exists( $key ) && &foo;
#
# See if a key exists
#
##################################################################

sub exists {
   my ( $self, $key ) = @_;

   return undef unless defined $self->{file};

   my @file = split /\n/, $self->{file};
   my $existance = 0;

   local $_;

LINE:
   foreach $_ ( @file ) {
      next LINE if /^#/; # comments start with ^#
      if ( /^$key((\!.*)|\s*)/ ) {
         $existance++;
      }
   }

   # match subkeys also i.e. modules!mime(!types...) exists

   return $existance;
}



##################################################################
#
# $zc->delete( $key )
#
# delete a key
#
##################################################################


sub delete {
   my ( $self, $key ) = @_;

   return undef unless defined $self->{file};

   # remove instances of this key on any line including the last one
   # subkey deletes entire branch

   $key = quotemeta( $key );

   my @file = split /\n/, $self->{file};

   for( my $i=0; $i<@file; $i++ ) {
      if ( ($file[$i] =~ m/^$key[\!\s]/) or ($file[$i] eq $key) ) {
         splice( @file, $i, 1 );
         $i--;
      }
   }

   $self->{file} = join "\n", @file;
   $self->{file} .= "\n" if( $self->{file} !~ /\n$/ );

   return 1;
}





##################################################################
# Get a filename for the temporary file
##################################################################

sub temporary_filename {
   my $path = shift @_;

   my @path = &split_path( $path );

   my $filename = pop @path;
   $filename = '.' . $filename;  # make the temporary file a dot file
   push @path, $filename;

   $path = &join_path( @path );

   my $temporary_filename = $path . ".tmp.$$";
   my $temporary_filename_base = $temporary_filename;
   my $i = 0;

   # ensure uniqueness

   while ( -e $temporary_filename ) {
      $i++;
      $temporary_filename = $temporary_filename_base . $i;
   }

   return $temporary_filename;
}


##################################################################
# Split a path into an array
##################################################################

sub split_path {
   my ($path) = @_;

   my @path = split /\//, $path;

   return @path;
}


##################################################################
# Join a path from a array
##################################################################

sub join_path {
   my @path = @_;

   my $path = join '/', @path;

   return $path;
}

1;

__END__


=head1 NAME

ZConf -  an interface to Zeus configuration files

=head1 SYNOPSIS

   use ZConf;

   # load a configuration file into memory

   $file = "$ENV{ZEUSHOME}/webadmin/conf/sites/server";
   $zc = new ZConf();
   $zc->load( $file );

   # Read values from the loaded configuration file

   $port = $zc->get('port');
   print "Port number is : $port\n";

   $htaccess = $zc->get('modules!htaccess!enabled');
   print "htaccess module enabled\n" if $htaccess eq 'yes';

   # Insert or change a key-value

   # add a new key/value pair
   $zc->add( 'my!new!key', '73' );
   # set the canonical hostname of this server to 'localhost'
   $zc->add( 'ip_name', 'localhost' );

   # Listing the next level of the key hierarchy
   # matching the given prefix

   @list = $zc->list('modules!mime!types');

   # Test whether a given key/branch exists
   # ( get will return "" for no value and when the key does not exist )

   $zc->exists( $key );

   # removing a key or branch

   $zc->delete( $key );

   # write the configuration file to disk

   $zc->save( $config_file );

   e.g. to add a mime type for .something files
        but only if the something module is enabled

   $zc = new ZConf;

   $zc->load( $config_file ) or die "Could not open configuration file";

   $val = $zc->get('module!something!enabled');

   if ( $val =~ /y/ ) {  # enabled
      $zc->add('modules!mime!types!something', 'application/x-something');
   }

   $val = $zc->save( $config_file );

   $zc = undef;

=head1 DESCRIPTION

This module provides an interface to easily access Zeus
configuration files. i.e. those in the same format as
$ZEUSHOME/webadmin/conf/sites/*

It is normally found in $ENV{ZEUSHOME}/admin/etc/ which you will
need to add to your @INC array if you are not using the miniperl
binary shipped with Zeus.

The Exporter module is not used so the full package names of methods
should be specified.

These files consist entirely of one line, whitespace separated
key-value pairs, although the value may be a null string.
The keys are arranged in a hierarchy with levels separated
by '!'.

e.g. modules!fastcgi!enabled     no

Where the 'modules' prefix indicates that this is associated
with a module. 'modules!fastcgi' indicates the FastCGI module.
The complete key 'modules!fastcgi!enabled' points to 'no',
the FastCGI module is not enabled.

If the first character of a line is '#' then it is treated as
a comment.

When reading a value from a configuration file you must specify
the filename and load that file into a ZConf object in memory.
The load method will return undef if the file cannot be loaded.

Use the get method to read values specifying the full key.
If the key is not present or the value is blank this will return
a null string. The object is not tied to the configuration file
so you will need to load the file again if the memory copy is old.

To add or modify values you must first make sure you have a fresh
copy of the file loaded. No locks are held so try to minimize the
time spent between loading the file and writing it back.
Use the add method to add or set values. The delete method can be
used to delete keys-value pairs.

Once modifications are complete overwrite the original file using the
save method.



=head2 Functions

=over 4

=item new ZConf([$filename]);

e.g. $zc = new ZConf("$ENV{ZEUSHOME}/admin/website");

Create a new ZConf object. Optionally set the filename of the configuration
file it is associated with. This is placed in $self->{filename}.
The file is not automatically loaded.

The new object will also store a copy of the configuration file in
$self->{file} once it has been loaded.

$self->{file} and $self->{filename} are internal values.


=item $zc->load([$filename]);

Load the file specified by $filename into an internal variable.
The contents of the file is also returned.
If this function fails the stored value is undefined and the
return value is undef.
The name of the last file loaded is remembered.

=item $zc->get($key);

Extract the value associated with this key. If the loaded file
is undefined this returns undef. If the key does not exist
or has no value a null string is returned. Otherwise the value
(The remainder of the line following the whitespace after the
key) is returned. If duplicate keys are present only the first
value is returned. Only complete key matches count. 'modules'
will not match 'modules!cgi'.


=item $zc->exists( $key );

If the key is found exactly or as a prefix returns true.
Otherwise returns false.
This is used in addition to get() as get does not differentiate
between a null value and a nonexistent key.


=item $zc->list( $prefix );

This is used to list the next level of a subtree of keys

  e.g. $zc->list( $prefix )

  will return

  access
  cgi
  dirlist ... etc

It will not return whole multilevel branches. Duplicate items will
be made unique. Exact matches are not counted.


=item $zc->add( $key, $value );

Used to set a key value pair. If the key already exists the value
is overwritten. Otherwise the pair is appended.

The changes made only affects the copy in memory.
The file must be saved for the changes to be committed.

Values cannot contain the "\n" new line character. If required it
should be encoded in some manner defined by the application using
the information.

Keys may not start with a '#' as this is the comment character at
the start of lines.

Only minimal changes are made to the file. It is not completely
reordered and comments are not stripped.

If the key is added successfully returns true, otherwise (e.g. due to
restrictions on allowable characters) returns false.


=item $zc->delete( $key );

Deletes all lines in a file matching the given key or key prefix.

=item $zc->flatten();

Returns a string containing the memory copy of the config file which
the object represents.

=item $zc->save( $filename );

Saves the memory copy of the file to the specified location or, by default,
to the previously loaded file's name.

This is done by writing a temporary file and moving it over the target file.

=back

=cut
