# SSL PKI routines for use by AHTML/CGI pages
#
# These routines (aside from new) return undef on success; on failure, they
# return a string describing the problem.  Typical calling:
#
# my $zs = new ZSSL( voice, clobber, verbose );
# my %fail = $zs->setDistinguishedName( hash );
# my $fail = $zs->checkPrivateKey(key);
# $fail = $zs->createPrivateKey(size, key) unless (-f key or $fail);
# $fail = $zs->createCSR(key, request) unless $fail;
# $fail = $zs->speakCSR([request]) unless $fail;
# $fail = $zs->checkCSR(request, cert) unless $fail;
# $fail = $zs->createSelfSignedCert(key, cert) unless $fail;
# $fail = $zs->readCert([cert]) unless $fail;
# $fail = $zs->checkCert(cert, key) unless $fail;
# $fail = $zs->speakPublicCert() unless $fail;
# my %hash = $zs->DNStringToHash('C=..., O=...');

package ZSSL;
my $cert = "$ENV{ZEUSHOME}/admin/bin/cert";

# $zs = new ZSSL(speakfunc, clobber, [gabbleful])
sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my $self = {};

    ($self->{speakfunc}, $self->{clobber}, $self->{verbose}) = @_;

    return bless $self, $class;
}

# $zs->setDistinguishedName( hash )
# Sets and checks DN
# Interesting fields of the hash:
# country, state, location, org, orgunit, common_name
# Returns undef if OK, otherwise hash from field to problem with field
sub setDistinguishedName {
    my $self = shift;
    my ( %args ) = @_;
    my ( %errors ) = ();
    my @interesting_keys = ( "country", "state", "location",
                             "org", "orgunit", "common_name"  );

    for $key ( @interesting_keys ) {
       $self->{$key} = $args{$key};
    }

    for $key ( 'common_name', 'org', 'location', 'country' ) {
       $errors{$key} = "Mandatory field" unless $self->{$key};
    }

    if ( $self->{country} =~ /US/i ) {
       $errors{"state"} = "Mandatory field for US" unless $self->{"state"};
    }

    $self->{distinguished} = 1 unless keys %errors;
    return %errors;
}

# $zs->checkPrivateKey([private])
#
# Verifies that the named file either doesn't exist or looks sensible.
sub checkPrivateKey {
    my $self = shift;
    my $name = shift || $self->{private};
    return 'No private key file specified' unless $name;
    return 'Use absolute path for private key, not ' . $name
        unless $name =~ m|^/|;

    if ( -f $name ) {
        return "Cannot read private key file, $name" unless -r _;
        my @s = stat _;
        my ($mode, $uid, $gid) = ($s[2], $s[4], $s[5]);
        return "All users can read your private key file, $name"
            if ($mode & 04) != 0; # world-readable

        $/ = "\n";
        open KEY, "<$name";
        my $first = <KEY>;
        close KEY;
        return "Private key file $name exists but is not a private key"
            unless $first =~ /^-+BEGIN RSA PRIVATE KEY-+\n$/;

        # Now check sensible group/user and prots.  We can read it, so
        # either $< owns the file, or $< is in the owning group.
        unless ( ($mode & 040) == 0 ) {
            my @group = getgrgid $gid;  # name, passwd, gid, "mem be rs"
            my @i = getpwnam $<;        # name, ...
            my @user = getpwnam $uid;
            my ($i, $u) = ($i[0], $user[0]);

            my @rivals = ();
            foreach ( split /\s+/, $group[3] ) {
                push @rivals, $_
                    unless ( $_ eq $i or $_ eq $u or $_ eq 'root' );
            }
            return "Private key file $name is readable by other users, e.g. "
                . join(', ', @rivals)
                    if @rivals;
            # group's members appear to be $u, $i and root: I guess we trust them.

            # $u *may* be whoever owns the site; otherwise, $u should be $i.
            $self->_speak( "\n<BR><B>Warning</B>: private key file $name is ",
                           ( $u eq $i ) ? 'group-readable' : "owned by $u, not $i",
                           ".\n" );
        }

        # Finally, check read-only (yes, this is anal).
        return "Private key file $name is not read-only"
            unless ($mode & 07333) == 0;
    }

    $self->{private} = $name;
    return undef;
}

# $zs->createPrivateKey([size, [filename]])
# Size defaults to 1024 if undef
#
# Creates a new private key.
sub createPrivateKey {
    my $self = shift;
    my ($size, $name) = (1024, $self->{private});
    
    $size = shift if @_;
    if ( @_ ) {
        $name = shift;
        return 'Use absolute path for private key file, not ' . $name
            unless $name =~ m|^/|;
    }
    return "Key file $name already exists: not over-written" if -f $name;

    my $stem = "$cert -new -type private";
    my $tail = "-keysize $size -out $name";

    $self->_speak( '<pre>' );
    $self->_speak( "$stem \\\n\t $tail\n\n" ) if $self->{verbose};

    $/ = "\n";
    open CMD, "$stem $tail 2>&1 |";
    while (<CMD>) { $self->_speak( "   $_" ); }
    $self->_speak( "</pre>\n" );

    close CMD or return 'Command exited abnormally - aborting';

    $self->_speak( "Your new private key has been written to `$name'\n" );
    $self->{private} = $name;
    $self->{private_created} = 1;

    return undef;
}

# $zs->_feed_id(stem, tail), for internal use, by ZSSL.pm, only
#
# Runs a command, feeding it distinguished name on stdin (as expected by
# $cert when given certain options).  The command run is "$stem $tail"
# but it will be displayed, in output, as "$stem \\\n\t $tail" to avoid
# over-long lines.
sub _feed_id {
    my $self = shift;
    return 'No distinguished name supplied' unless $self->{distinguished};
    my ($stem, $tail) = @_;
    $self->_speak( '<pre>' );
    $self->_speak( "$stem \\\n\t $tail\n" ) if $self->{verbose};
    $tempfile = _getTmpName();

    open CMD, "|$stem $tail 2>&1 > $tempfile";
    # feed %id in on stdin (in expected order)

    @keys = ( "country", "state", "location", "org", "orgunit", "common_name"  );
    for $key ( @keys ) {
       if ( $self->{$key} ) {
          print CMD $self->{$key};
       } else {
          print CMD ".";
       }
       print CMD "\n";
    }
    close CMD;
    if ($?) {
        $self->_speak( "</pre>\n" );
        unlink $tempfile;
        return 'Command exited abnormally - aborting';
    }

    my $txt;
    if ( $self->{speakfunc} ) {
        local $/;
        undef $/;
        open IN, $tempfile;
        $txt = "   " . <IN>;
        close IN;
    }
    unlink $tempfile;

    if ( $self->{speakfunc} ) {
        # The output only shows the prompts we would have been offered had
        # this been interactive: fill in the details we just piped in on stdin.
        $txt =~ s|:  |: <b>$self->{country}</b>\n|;
        $txt =~ s|:  |: <b>$self->{state}</b>\n|;
        $txt =~ s|:  |: <b>$self->{location}</b>\n|;
        $txt =~ s|:  |: <b>$self->{org}</b>\n|;
        $txt =~ s|:  |: <b>$self->{orgunit}</b>\n|;
        $txt =~ s|:  |: <b>$self->{common_name}</b>\n|;
        $txt =~ s|\n|\n   |g; #indent
        $self->_speak( "$txt</pre>\n" );
    }
    return undef;
}

sub _run_cmd {
    my ($self, $stem, $tail) = @_;

    $tail = "" unless defined( $tail );

    if ( $self->{verbose} ) {
        $self->_speak( "<PRE>$stem" );
        $self->_speak( " \\\n\t $tail" ) if $tail;
        $self->_speak( "</PRE>\n" );
    }
    $self->_speak( "<PRE>\n" );
    $/ = "\n";

    open CMD, "$stem $tail 2>&1 |";
    while (<CMD>) { $self->_speak( Z::quote_html($_) ); }
  
    $self->_speak( "</PRE>\n" );
    close CMD or return 'Command exited abnormally - aborting';

    return undef;
}

# $zs->createCSR( private_key_filename, csr_filename )
# private_key_filename can be undef if you have defined
# private elsewhere
#
# Stores a certificate request in the named file.
sub createCSR {
    my $self = shift;
    my ( $private_name, $csr_name ) = @_;
    $private_name = $self->{private} if ( !defined $private_name );
    $csr_name = $self->{request} if ( !defined $csr_name );

    return 'Use absolute path for certificate request file, not ' . $csr_name
        unless $csr_name =~ m|^/|;
    return 'Cannot generate certificate request until private key has been specified'
        unless $private_name;

    if ( $self->{clobber} or ! -f $csr_name ) {
        unlink $csr_name if -f $csr_name;
        my $fail =
            $self->_feed_id( "$cert -new -type request",
                             "-key $private_name -out $csr_name" );
        return $fail if $fail;

        $self->_speak( "Wrote certificate request to $csr_name\n" );

    } else {
        my $found = "$csr_name already exists";
        return "$found: please move or delete it and try again"
            if $self->{private_created};

        $self->_speak( "$found: assuming it matches private key in $self->{private}.\n",
                       "If not, please move or delete it and try again.\n" );
    }

    $self->{request} = $csr_name;
    return undef;
}

# $zs->speakCSR([filename])
#
# Displays contents of $zs->{request} file.
sub speakCSR {
    my $self = shift;
    my $name = shift || $self->{request};
    return $self->_run_cmd( "$cert -in $name -text" );
}

# $zs->checkCSR([CSR, [cert]])
#
# Returns undef if the CSR and cert match; otherwise, returns an HTML string 
# describing the problem, if any; if there are several, this will be of form
# '<UL><LI>...<LI>...</UL>' with suitable internal (but no trailing) newlines.
sub checkCSR {
    my ($self, $csr, $cert) = @_;
    $csr = $self->{request} unless $csr;
    $cert = $self->{public} unless $cert;

    my %request = $self->readCert($csr);
    my %public;
    if ( $cert ) {
        %public = $self->readCert($cert);
    } else {
        # generate temp cert from self's private key (tedious, but hey).
        my $aid = new ZSSL(); # terse and gentle
        my %err = $aid->setDistinguishedName(%{$self});
        unless ( keys %err ) {
            $aid->createSelfSignedCert($self->{private}, $self->_getTmpName());
            %public = $self->readCert($aid->{public});
            unlink $aid->{public};
        }
    }

    my @faults = ();
    if ( keys %public ) {
        my @miss = ();

        foreach ( keys %{$request{subject}} ) {
            push @miss, $_ unless exists $public{subject}->{$_};
        }
        push @faults, "CSR DN lacks @miss" if @miss;
        @miss = ();
        my @diff = ();

        while ( my ($k, $v) = each %{$public{subject}} ) {
            if ( exists $request{subject}->{$k} ) {
                push @diff, $k unless $v eq $request{subject}->{$k};
            } else {
                push @miss, $k;
            }
        }
        push @faults, "Key DN lacks @miss" if @miss;
        push @faults, "Subject distinguished names differ in @diff" if @diff;
        push @faults,
        "Moduli do not match: <PRE>\n$request{modulus}\n$public{modulus}</PRE>"
            if $request{modulus} ne $public{modulus};

    } else {
        push @faults, "Unable to check CSR: no distinguished name data available";
    }

    return undef unless @faults;
    return $faults[0] unless $#faults > 1;
    return "<UL><LI>" . join("\n<LI>", @faults) . "</UL>";
}
# $zs->createSelfSignedCert( private_key_filename, cert_filename )
# private_key_filename can be undef if you have defined private
# somewhere else
#
# Generates a (self-signed) certificate of the public key.
sub createSelfSignedCert {
    my $self = shift;
    my ( $private_name, $cert_name ) = @_;
    $private_name = $self->{private} if ( !defined $private_name );
    $cert_name = $self->{public} if ( !defined $cert_name );

    return 'Use absolute path for public certificate, not ' . $cert_name
        unless $cert_name =~ m|^/|;

    if ( $self->{clobber} or ! -f $cert_name ) {
        unlink $cert_name if -f $cert_name;
        my $fail = $self->_feed_id( "$cert -new -type public",
                                    "-key $private_name -out $cert_name" );
        return $fail if $fail;

        $self->_speak( "Your new public key has been written to `$cert_name'.\n" );

    } else {
        my $found = "$cert_name already exists";
        return "$found: please move or delete it and try again.\n"
            if $self->{private_created};

        $self->_speak( "$found: assuming this matches private key in $private_name.\n",
                       "If not, please move or delete it and try again.\n" );
    }

    $self->{public} = $name;
    return undef;
}

# $zs->readCert([filename])
# Reads data in from a certificate (or request) file, returns a hash containing:
# modulus -> string, matches /^([0-9a-fA-F][0-9a-fA-F]:)+$/
# subject -> subjectHash
# and, for an actual certificate:
# issuer -> issuerHash
# self_signed -> 0/1
# valid_from_text
# valid_until_text
# serialno_text -> string
# (the two *Hashes are compatible with the input of createCSR)

sub readCert
{
    my $self = shift;
    my $file = shift || $self->{public};
    open CERTFILE, "$cert -text -in $file |";

    my %result = ();
    my $reading = ''; # one of: 'dn', 'mod', 'validity'
    my $opening_whitespace = '';

    while ( <CERTFILE> ) {
        if ( /^(\s*)Subject:/ ) {
            $subject_dn = "";
            $dnptr = \$subject_dn;
            $reading = 'dn';
            $opening_whitespace = $1;
            next;

        } elsif ( /^(\s*)Issuer:/ ) {
            $issuer_dn = "";
            $dnptr = \$issuer_dn;
            $reading = 'dn';
            $opening_whitespace = $1;
            next;

        } elsif ( /^(\s*)Modulus:/ ) {
            $modulus = '';
            $reading = 'mod';
            $opening_whitespace = $1;
            $_ = $'; # in case older version of cert with missing \n
            next unless /\S/;

        } elsif ( /^(\s*)Serial Number:/ ) {
            $result{"serialno_text"} = '';
            $reading = 'serial';
            $opening_whitespace = $1;
            $_ = $'; # in case older version of cert with missing \n
            next unless /\S/;

        } elsif ( /Validity:/ ) {
            $reading = 'validity';
            next;
        }

        # If we're reading a block, has it ended ?
        # Detect by indentation level dropping back to same as when we started:
        if ( /^$opening_whitespace\S/ ) {
            $reading = '';
            $opening_whitespace = '';
        }

        if ( $reading eq 'validity' ) {
            if ( /Not After:\s*(.*)$/ ) {
                $result{"valid_until_text"} = $1;
                next;
            }
            if ( /Not Before:\s*(.*)$/ ) {
                $result{"valid_from_text"} = $1;
                next;
            }

        } elsif ( $reading eq 'mod' ) {
            /^\s*([0-9a-fA-F:]*)\s*$/;
            $modulus .= $1;

        } elsif ( $reading eq 'serial' ) {
            /^\s*(\S+)\s*$/;
            $result{"serialno_text"} .= $1;

        } elsif ( $reading eq 'dn' ) {
            # Join into one long DN string
            $$dnptr .= ", " . $_;
            next;
        }
    }

    close( CERTFILE );

    my %iss = $self->DNStringToHash( $issuer_dn );
    my %sub = $self->DNStringToHash( $subject_dn );
    $result{"issuer"} = \%iss if keys %iss;
    $result{"subject"} = \%sub if keys %sub;

    $result{"self_signed"} = ( $issuer_dn eq $subject_dn );
    # modulus is a string; pairs of hex digits, separated by colons.
    $result{modulus} = $modulus; # key-size is (1+length)/3 bytes
    return %result;
}

# DNStringToHash('C=..., O=...') - Converts a distinguished name to a hash.
sub DNStringToHash {
   my ($self, $dn) = @_;
   my %hash_tags = (
                    "C" => "country",
                    "ST" => "state",
                    "L" => "location",
                    "O" => "org",
                    "OU" => "orgunit",
                    "CN" => "common_name"
                   );

   my %retval;
   for $key (keys %hash_tags) {
      $dn =~ /$key=(.*?)($|(,\s*[A-Z]+=))/s;
      my $val = $1;
      $val = "" unless defined( $val );
      $val =~ s/\s+$//;
      $retval{$hash_tags{$key}} = $val;
   }

   return %retval;
}

# $zs->checkCert([public, [private]])
#
# Checks that a certificate matches a private key.
sub checkCert {
    my $self = shift;
    my $publ = shift || $self->{public};
    my $priv = shift || $self->{private};
    my $fail = $self->_run_cmd( "$cert -check", "-key $priv -in $publ" );
    return $fail . ' - keys probably do not match' if $fail;

    if ( $self->{distinguished} ) {
        my %cert = $self->readCert($publ);
        my @diff = ();
        while ( my ($k, $v) = each %{$cert{subject}} ) {
            push @diff, "$k ($v)" unless $self->{$k} eq $v;
        }
        return "Certificate conflicts with distinguished name in " . join(', ', @diff)
            if @diff;
    }
    return undef;
}

# $zs->speakPublicCert([filename])
#
# Displays the certificate held by the named file.
sub speakPublicCert {
    my $self = shift;
    my $name = shift || $self->{public};
    return $self->_run_cmd( "$cert -in $name -text" );
}

sub readPrivateKey
{
    my $self = shift;
    my $file = shift || $self->{private};
    open KEYFILE, "$cert -text -in $file |";

    my %result = ();
    my $reading = '';
    my $opening_whitespace = '';

    while ( <KEYFILE> ) {
       chomp;
       if ( /^(\s+)([^\s:]+):\s*$/ ) {
          # It's a multi-line item
          $opening_whitespace = $1;
          $reading = $2;
          $result{$2} = "";
          $multiline_ptr = \$result{$2};
          next;
       } 
       
       # If we're reading a block, has it ended ?
       # Detect by indentation level dropping back to same as when we started:
       if ( /^$opening_whitespace\S/ ) {
          $reading = '';
          $opening_whitespace = '';
       } elsif ( defined( $multiline_ptr ) ) {
          $_ =~ s@^\s*@@; # Strip leading whitespace
          $_ =~ s@\s*$@@; # Strip trailing whitespace

          # Join into one long string
          $$multiline_ptr .=  $_;
          next;
       }
    }
    
    close( CERTFILE );
    
    $result{"key_size"} = 8 * ((length($result{"n"})+1)/3);

    return %result;
}

# routine to speak to stuff
sub _speak
{
   my ( $self, @outputs ) = @_;
   my $func;

   $func = $self->{speakfunc};

   if ( defined( $func ) ) {
      &$func(@outputs);
   }
}

# Routine to return a safe temporary filename
sub _getTmpName
{
   return "$ENV{ZEUSHOME}/admin/internal/cert-zssl.$$";
}

1;
